/*!
* \file obj_allocator.hpp
* \brief Automatic OBJ VRAM allocations
 */

#pragma once

#include "AstralbrewEngine/memory/address.hpp"

namespace Astralbrew::Allocator::ObjAllocator
{
	/*!
	* \brief Allocates a block of the specified size and exposes its address.
	* 
	* \param [out] address The movable address containing the allocated OBJ VRAM pointer
	* \param [in] blocks_magnitude Size of the allocated object block.	
	* 
	* \details For the blocks magnitude, use the `BLOCKS_SPR_<W>x<H>_<D>bit` constants, where
	* W, H are the object width, height (8,16,32,64), and D is the palette bit depth (4,8).
	 */
	void alloc(Memory::Address* address, int blocks_magnitude);
	
	/*!
	* \brief Frees the OBJ VRAM block allocated at the address
	* 
	* \param [in, out] address address of allocated block	
	* 
	* \details If the deallocation succeeds, the value of the address is set to NULL. Freeing 
	* a previously unallocated or out of range address has no effect through this function call.
	 */
	void dealloc(Memory::Address* address);	
	
	extern const int BLOCKS_SPR_8x8_4bit;  /*!< Block magnitude of a 4-bit 8x8 object */
	extern const int BLOCKS_SPR_8x16_4bit; /*!< Block magnitude of a 4-bit 8x16 object */
	extern const int BLOCKS_SPR_16x8_4bit; /*!< Block magnitude of a 4-bit 16x8 object */
	extern const int BLOCKS_SPR_16x16_4bit; /*!< Block magnitude of a 4-bit 16x16 object */
	extern const int BLOCKS_SPR_16x32_4bit; /*!< Block magnitude of a 4-bit 16x32 object */
	extern const int BLOCKS_SPR_32x16_4bit; /*!< Block magnitude of a 4-bit 32x16 object */
	extern const int BLOCKS_SPR_32x32_4bit; /*!< Block magnitude of a 4-bit 32x32 object */
	extern const int BLOCKS_SPR_32x64_4bit; /*!< Block magnitude of a 4-bit 32x64 object */
	extern const int BLOCKS_SPR_64x32_4bit; /*!< Block magnitude of a 4-bit 64x32 object */
	extern const int BLOCKS_SPR_64x64_4bit; /*!< Block magnitude of a 4-bit 64x64 object */
	
	extern const int BLOCKS_SPR_8x8_8bit;  /*!< Block magnitude of an 8-bit 8x8 object */
	extern const int BLOCKS_SPR_8x16_8bit; /*!< Block magnitude of an 8-bit 8x16 object */
	extern const int BLOCKS_SPR_16x8_8bit; /*!< Block magnitude of an 8-bit 16x8 object */
	extern const int BLOCKS_SPR_16x16_8bit; /*!< Block magnitude of an 8-bit 16x16 object */
	extern const int BLOCKS_SPR_16x32_8bit; /*!< Block magnitude of an 8-bit 16x32 object */
	extern const int BLOCKS_SPR_32x16_8bit; /*!< Block magnitude of an 8-bit 32x16 object */
	extern const int BLOCKS_SPR_32x32_8bit; /*!< Block magnitude of an 8-bit 32x32 object */
	extern const int BLOCKS_SPR_32x68_8bit; /*!< Block magnitude of an 8-bit 32x68 object */
	extern const int BLOCKS_SPR_68x32_8bit; /*!< Block magnitude of an 8-bit 68x32 object */
	extern const int BLOCKS_SPR_68x68_8bit; /*!< Block magnitude of an 8-bit 68x68 object */
}