/*!
* \file callstack.hpp
* \brief Call stack debugging markers definitions
 */

#pragma once

#include "AstralbrewEngine/debug/error.hpp"

/*! \def callstack_call
* \brief Macro that puts the current function on to the call stack
* 
* \details Should be called at the beginning of the "stackable" function.
* 	During the execution, each callstack_call must be paired with a callstack_ret,
* 	otherwise the call stack will cause overflowing issues.
 */
 
#define callstack_call \
	if(Astralbrew::Debug::callstack_size == CALLSTACK_MAX_SIZE) \
		Astralbrew::Debug::fatal_error("Callstack size exceeded");	\
	Astralbrew::Debug::callstack_file_name[Astralbrew::Debug::callstack_size] = __FILENAME__; \
	Astralbrew::Debug::callstack_func_name[Astralbrew::Debug::callstack_size++] = __PRETTY_FUNCTION__;

/*! \def callstack_ret
* \brief Macro that pops the current function from the call stack. Replaces the `return` keyword with some extra precautions.
* 
* \details Do not callstack_ret in a function that doesn't do a callstack_call before. Surround callstack_ret between between braces {}
* when this is the only instruction in one of the if statement branches.
* 
* \code{.cpp}
* int f(int x)
* {
* 	callstack_call
* 	if(x % 2 == 1)
* 		FATAL_ERROR("lalala"); // this triggers call stack to be printed out on the screen
* 	else 
* 	{
* 		callstack_ret x/2; // note the braces around this line
* 	} 
* }
* \endcode
 */

 
#define callstack_ret \
	if(Astralbrew::Debug::callstack_size == 0) \
		Astralbrew::Debug::fatal_error("Callstack empty");	\
	Astralbrew::Debug::callstack_size--; \
	return
	

/*! \cond HIDDEN_SYMBOLS */ 

#define CALLSTACK_MAX_SIZE 128

namespace Astralbrew::Debug
{	
	__attribute__((section(".ewram")))	
	extern int callstack_size;
	
	__attribute__((section(".ewram")))	
	extern const char* callstack_file_name[CALLSTACK_MAX_SIZE];
	
	__attribute__((section(".ewram")))	
	extern const char* callstack_func_name[CALLSTACK_MAX_SIZE];	
}

/*! \endcond */