/*!
* \file class.hpp
* \brief Object class definitions
 */

#pragma once

#include "AstralbrewEngine/doxyfile_metadata.hpp"

/*!
*
* \brief Object class type
*/
typedef unsigned short obj_class_t;

/*! \cond HIDDEN_SYMBOLS */ 
#define AT_GLOBAL_SCOPE (__builtin_FUNCTION()[0] == '\0')

/*! \endcond */ 

/*!
*
* \brief Refers to an object class
* 
* \param [in] name Object class name
* \return Object class id
* 
* 
* \details Useful when different object types produce different behaviors
* \code{.cpp}
* if(enemy->is_of_class(class_of(Alligator)))
* {
* 	player->run();
* }
* \endcode
* 
* \details Shortcut `$ (name)` can also be used: `enemy->is_of_class($ (Alligator))`.
*
 */
#define class_of(name) __obj_class_##name

/*! \cond HIDDEN_SYMBOLS */ 
#define $(name) class_of(name)
/*! \endcond */

//* \details This macro can only be called **globally**.

/*!
*
* \brief Creates an object class
* \param [in] name object class name
* 
* \details This macro must be called **globally**.
* \code{.cpp}
* register_class(Alligator);
* 
* int main() { ... }
* \endcode
* 
* \details There is a `no_class` record used as default class name for object which have not been assigned a class.
* 
* \code{.cpp}
* if(sprite->is_of_class(class_of(no_class))) { ... }
* \endcode
 */
#define register_class(name) const obj_class_t $(name) { AT_GLOBAL_SCOPE ? __COUNTER__ : -1};
// will raise error on __COUNTER__ ushort overflo------^
// prevent register_class definitions to be made locally

register_class(no_class); // let 0 be the "default" class value





