/*!
* \file obj_visual.hpp
* \brief Object Visual definition
 */

#pragma once

#include <gba.h>

#include "AstralbrewEngine/types/fixed_vector.hpp"
#include "AstralbrewEngine/types/vector.hpp"
#include "AstralbrewEngine/memory/address.hpp"

namespace Astralbrew::Entity
{
	/*! \brief OAM Object graphics handler class	
	 */
	class ObjVisual
	{
		private:
			Vector<const Memory::Address*> frames;
			u8 crt_gfx_id = 0;
			
			FixedVector<FixedVector<u8,8>, 8> framesets;			
			// bit  0    : enabled [0..1]
			// bit  1    : loop    [0..1]
			// bit  2-4  : frameset id [0..7]
			// bit  5-7  : frame id inside frameset [0..7]
			u8 anim_stat = 0;
			// bit  0-3 : anim ticks
			// bit  4-8 : current ticks
			u16 tick_stat = 0;
			
		public:
			/*!
			* \brief Creates an object visual data with a certain number of frames			
			* \param [in] frames_count number of reserved frames for sprite
			 */
			ObjVisual(u16 frames_count);
			
			/*!
			* \brief Gets number of frames			
			* \return number of reserved frames for sprite			
			 */
			u8 get_frames_count() const;
			
			/*!
			* \brief Assigns a frame to some graphics data
			* 
			* \param [in] id Frame id
			* \param [in] obj_tile_address Address of the object tiles			
			 */
			void set_frame(int id, const Astralbrew::Memory::Address* obj_tile_address);

			/*!
			* \brief Makes animation run one frame at a number of ticks
			* 
			* \param [in] ticks Ticks until the next frame			
			 */
			void set_animation_ticks(u8 ticks);
			
			/*!
			* \brief Sets animation frames into one of the 8 frame slots
			* 
			* \param [in] frameset_id Animation frame slot [0..7]
			* \param [in] ... List of frame ids [0..frames count]			
			 */
			void set_animation_frames(u8 frameset_id, ...);
			
			
			/*!
			* \brief Makes animation run one of the 8 frame slots
			* 
			* \param [in] frameset_id Animation frame slot [0..7]
			* \param [in] loop Animation loop			
			 */
			void set_animation_track(u8 frameset_id, bool loop);
			
			/*!
			* \brief Gets the object current displayed graphics
			* 
			* \return Address of the cuurent displayed graphics data			
			 */			
			const Memory::Address* get_crt_gfx() const;			
			
			/*!
			* \brief Sets the object current displayed graphics
			* 
			* \param [in] frame_id Frame number to get the current graphics from
			* 			
			* \details **Caution!** Calling set_crt_gfx() disables the animation.
			 */
			void set_crt_gfx(int frame_id);				
			
			/*!
			* \brief Updates graphics data (run animation etc.)					
			 */
			void update();
	};
	
	
	extern const int ANIM_FRAMES_0; /*!< Literal name for animation frame 0*/
	extern const int ANIM_FRAMES_1; /*!< Literal name for animation frame 1*/
	extern const int ANIM_FRAMES_2; /*!< Literal name for animation frame 2*/
	extern const int ANIM_FRAMES_3; /*!< Literal name for animation frame 3*/
	extern const int ANIM_FRAMES_4; /*!< Literal name for animation frame 4*/
	extern const int ANIM_FRAMES_5; /*!< Literal name for animation frame 5*/
	extern const int ANIM_FRAMES_6; /*!< Literal name for animation frame 6*/
	extern const int ANIM_FRAMES_7; /*!< Literal name for animation frame 7*/
}



