/*!
* \file sprite.hpp
* \brief Sprite definitions
 */

#pragma once

#include "AstralbrewEngine/objects/oam.hpp"
#include "AstralbrewEngine/memory/address.hpp"
#include "AstralbrewEngine/entity/class.hpp"
#include "AstralbrewEngine/entity/hitbox.hpp"
#include "AstralbrewEngine/entity/localizable.hpp"
#include "AstralbrewEngine/entity/obj_visual.hpp"
#include "AstralbrewEngine/entity/hitbox.hpp"
#include "AstralbrewEngine/types/point.hpp"
#include "AstralbrewEngine/world/camera.hpp"

namespace Astralbrew::Entity
{		
	/*! \brief Simple sprite object	
	 */
	class Sprite : public Localizable
	{
	private:		
		obj_class_t _class;
		Objects::ObjAttribute* _attr;			
						
		// internal layout
		Hitbox hitbox;
		Point<u8> anchor;
		Point<u8> theoretical_anchor;
		
		Point<sf24> pos;
		
		ObjVisual* visual;
		
		void swap_oam_pos(Sprite* other);
	public:					
		/*!
		* \brief Creates sprite object
		* 
		* \param [in] size Sprite size
		* \param [in] bit_depth Palette bit depth (4/8 bit)
		* \param [in] frames_count Number of reserved frames
		* \param [in] obj_class Object class				
		 */
		Sprite(Objects::ObjSize size, 
			Astralbrew::Objects::ObjBitDepth bit_depth, 
			u16 frames_count = 1, 
			obj_class_t obj_class = $(no_class));
		
		/*!
		* \brief Gets sprite internal object attribute		
		* \return Object attribute		
		* \details **Caution!** Do not attempt to free this pointer. Do not call OamPool::remove_obj() on this pointer.
		 */
		Astralbrew::Objects::ObjAttribute* get_attribute();
		
		/*! \brief get sprite hitbox
			\return hitbox coordinates
		*/
		Hitbox get_hitbox() const;

		virtual s16 pos_x() const override;
		virtual s16 pos_y() const override;

		/*!
		* \brief Sets sprite coordinates
		* 
		* \param [in] x Coordinate x
		* \param [in] y Coordinate y
		 */
		void set_position(sf24 x, sf24 y);
		
		/*!
		* \brief Sets sprite anchor, scaled 0..255 (128,128 = sprite center)
		* 
		* \param [in] x Anchor x
		* \param [in] y Anchor Y		
		* 
		* \details You can use shortcut definitions for trivial anchors: ANCHOR_TOP_LEFT etc
		 */
		void set_anchor(u8 x, u8 y);
		
		/*!
		* \brief Sets sprite anchor from point, scaled 0..255 (128,128 = sprite center)
		* 
		* \param [in] anchor Sprite anchor				
		 */
		void set_anchor(const Point<u8>& anchor);
		
		/*!
		* \brief Sets sprite hitbox		
		* \param [in] _hitbox Sprite hitbox		
		 */
		void set_hitbox(const Hitbox& _hitbox);		
		
		/*!
		* \brief Gets object visual data
		* 
		* \return Sprite object visual data
		* 
		* \details **Caution!** Do not attempt to free this pointer!
		 */
		ObjVisual* get_visual() const;
		
		/*!
		* \brief Updates sprite visual data		
		 */		
		void update_visual();
		
		/*!
		* \brief Updates sprite on screen position relative to a Camera
		* 
		* \param [in] cam World camera, NULL if there is no camera		
		 */
		void update_position(const World::Camera* cam);
		
		/*!		
		* \return Sprite top coordinate		
		 */
		sf24 get_top_coord() const;
		
		/*!		
		* \return Sprite left coordinate		
		 */
		sf24 get_left_coord() const;
		
		/*!		
		* \return Sprite right coordinate		
		 */
		sf24 get_right_coord() const;
		
		/*!		
		* \return Sprite bottom coordinate		
		 */
		sf24 get_bottom_coord() const;
		
		/*!
		* \brief Checks if sprite is of specified class
		* 
		* \param [in] xclass target object class
		* \return True if sprite is of the specified class, false otherwise		
		 */
		bool is_of_class(obj_class_t xclass) const;
		
		/*!
		* \brief Check if no class has been assigned to this sprite
		* 
		* \return True if sprite has no class, halse otherwise		
		 */
		bool has_no_class() const;
		
		/*!
		* \brief Check hitbox collision between two sprites
		* 
		* \param [in] spr Sprite to test collision with
		* \return True if the two sprites touch, false otherwise		
		 */
		bool touches(const Sprite* spr) const;
		
		virtual ~Sprite();
		
		/*!
		* \brief Create a one-frame 4-bit Sprite
		* 
		* \param [in] obj_tile Address to the graphics data
		* \param [in] size Sprite size
		* \param [in] anch_x Sprite anchor x
		* \param [in] anch_y Sprite anchor y
		* \return Sprite with the given properties		
		 */
		static Sprite* quick16(const Memory::Address* obj_tile, Objects::ObjSize size, u8 anch_x, u8 anch_y);
		
		/*!
		* \brief Create a one-frame 8-bit Sprite
		* 
		* \param [in] obj_tile Address to the graphics data
		* \param [in] size Sprite size
		* \param [in] anch_x Sprite anchor x
		* \param [in] anch_y Sprite anchor y
		* \return Sprite with the given properties		
		 */
		static Sprite* quick256(const Memory::Address* obj_tile, Objects::ObjSize size, u8 anch_x, u8 anch_y);
		
		void show_in_front_of(Sprite* other);
	};
	
	// Sprite anchor shortcut definitions

	/*           Name                X    Y  */	
	#define ANCHOR_TOP_LEFT           0,   0
	#define ANCHOR_TOP              128,   0
	#define ANCHOR_TOP_RIGHT        255,   0
	#define ANCHOR_LEFT               0, 128
	#define ANCHOR_CENTER           128, 128
	#define ANCHOR_RIGHT            255, 128
	#define ANCHOR_BOTTOM_LEFT        0, 255
	#define ANCHOR_BOTTOM           128, 255
	#define ANCHOR_BOTTOM_RIGHT     255, 255

	// Usage : sprite->set_anchor(ANCHOR_CENTER)
}

















