/*!
* \file address.hpp
* \brief Movable address type definition
 */
 
#pragma once

namespace Astralbrew::Memory
{
	class Address
	{
	private:		
		unsigned int value = 0;
		
	public:
		/*!
		* \brief Gets raw pointer address value
		* 
		* \return Pointer of address
		* 
		* \details The raw pointer beind the address may change during its lifetime. 
		* 	Use has_changed() and get_value() every time you need the effective address.
		* \details If you only need to access _what is_ at that address, use read() and write()
		*   methods instead.
		 */
		void* get_value() const;	
		
		/*!
		* \brief Sets the raw value of the address
		* 
		* \param [in] addr New pointer value
		 */
		void set_value(void* addr);
		
		/*!
		* \brief Checks if the address has changed
		* 
		* \return True if the address has changed, false otherwise
		* 
		* \details After this function call, the address "changed" flag is reset.
		* Do **not** do this:
		* \code{.cpp}
		* Address* address; 
		* address->set_value(0x02000000); //  change the address
		* 
		* if(condition1 && address->has_changed()) // has_changed() returns TRUE, changed flag reset
		* 	{ do_X(); }
		* if(condition2 && !address->has_changed()) // has_changed() returns FALSE!
		* 	{ do_Y(); }
		* \endcode
		* 
		* Instead, do this:
		* \code{.cpp}
		* bool address_changed = address->has_changed(); 
		* 
		* // good to go
		* if(condition1 && address_changed)
		* 	{ do_X(); }
		* if(condition2 && !address_changed)
		* 	{ do_Y(); }
		* 
		* \endcode
		* 
		 */
		bool has_changed();
		
		/*!
		* \brief Adds a scalar to the scalat
		* 
		* \param [in] offset Scalar value
		* \return New address = Old address + offset		
		 */
		Address operator + (int offset);
		
		/*!
		* \brief Subtracts a scalar from the scalat
		* 
		* \param [in] offset Scalar value
		* \return New address = Old address - offset		
		 */
		Address operator - (int offset);
		
		/*!
		* \brief Writes bytes from a source to this address
		* 
		* \param [in] source Source buffer
		* \param [in] size Number of bytes to copy		
		 */
		void write(const void* source, unsigned int size) const;
		
		/*!
		* \brief Reads bytes from this address
		* 
		* \param [in] dest Destination buffer
		* \param [in] size Number of bytes to copy		
		 */
		void read(void* dest, unsigned int size) const;
		
		/*!
		* \brief Checks if the raw pointer in in VRAM memory region		
		* \return True if address is in the VRAM, false otherwise		
		 */
		bool is_vram() const;
		
		/*!
		* \brief Checks if the raw pointer in in internal WRAM memory region		
		* \return True if address is in the internal WRAM, false otherwise		
		 */
		bool is_iwram() const;
		
		/*!
		* \brief Checks if the raw pointer in in external WRAM memory region		
		* \return True if address is in the external WRAM, false otherwise		
		 */
		bool is_ewram() const;		
		
		/*!
		* \brief Map the memory into a tileset starting from a certain address and
		*     tells the number of the tile the curent address belongs to
		* 
		* \param [in] head Begin of the tileset
		* \param [in] is_8bit True if tiles are 8-bit (64 bytes/tile), false if 4-bit (32 bytes/tile)
		* \return Id of the tile containing the current address		
		 */
		int get_tile_number(const void* head,  bool is_8bit) const;		
	};
}