/*!
* \file oam.hpp
* \brief Objects handler API
 */

#pragma once

#include <gba.h>

namespace Astralbrew::Objects
{
	
	/*! \enum ObjSize
		\brief all possible GBA sprite sizes
	 */
	enum ObjSize 
	{
		SIZE_8x8   =  0, //!< Sprite size 8x8  
		SIZE_16x16 =  1, //!< Sprite size 16x16 
		SIZE_32x32 =  2, //!< Sprite size 32x32 
		SIZE_64x64 =  3, //!< Sprite size 64x64 
		SIZE_16x8  =  4, //!< Sprite size 16x8  
		SIZE_32x8  =  5, //!< Sprite size 32x8  
		SIZE_32x16 =  6, //!< Sprite size 32x16 
		SIZE_64x32 =  7, //!< Sprite size 64x32 
		SIZE_8x16  =  8, //!< Sprite size 8x16  
		SIZE_8x32  =  9, //!< Sprite size 8x32  
		SIZE_16x32 = 10, //!< Sprite size 16x32 
		SIZE_32x64 = 11  //!< Sprite size 32x64 
	};
	
	/*! \enum ObjBitDepth
		\brief object bit depth (4-bit/8-bit)
	 */
	enum ObjBitDepth
	{
		_4bit = 0, //!< 16 colors Sprite
		_8bit = 1  //!< 256 colors Sprite
	};		
	
	/*! \class ObjAttribute
		\brief describes OAM object attributes, works only with 256-color sprites	
	 */
	class ObjAttribute
	{
	private:
		u16 attr0, attr1, attr2, attr3;
		ObjAttribute() = default;
	public:				
		/*! \brief creates an object attribute using 8-bit palette
			\param size       object size
			\param x          object position - horizontal coordinate
			\param y          object position - vertical coordinate
			\param tile_index tile number in OAM Tiles section of VRAM
			\param hflip      horizontal flip (0 = no ; non-0 = yes)
			\param hflip      vertical flip (0 = no ; non-0 = yes)
			\param priority   object priority (0..3)
		 */ 
		static ObjAttribute create256(ObjSize size, s16 x=0, s16 y=0, u16 tile_index=0, u16 hflip=0, u16 vflip=0, u16 priority=0);
		
		/*! \brief creates an object attribute using 4-bit palette
			\param size         object size
			\param x            object position - horizontal coordinate
			\param y            object position - vertical coordinate
			\param tile_index   tile number in OAM Tiles section of VRAM
			\param palette_bank palette bank (0..15)
			\param hflip        horizontal flip (0 = no ; non-0 = yes)
			\param hflip        vertical flip (0 = no ; non-0 = yes)
			\param priority     object priority (0..3)
		 */ 
		static ObjAttribute create16(ObjSize size, s16 x=0, s16 y=0, u16 tile_index=0, u16 palette_bank=0, u16 hflip=0, u16 vflip=0, u16 priority=0);
		
		/*! \brief set object size
			\param size object size
		 */
		void set_size(ObjSize size);
		
		/*! \brief get object size		
			\return object size 
		 */
		ObjSize get_size() const;
		
		/*! \brief flip object vertically 	
		 */
		void flip_v();
		
		/*! \brief set object vertical flip
			\param flip 0 if no flip, non-0 value if flipped
		 */
		void set_flip_v(u16 flip);
		
		/*! \brief get object vertical flip
			\return 0 if no flip, 1 if flipped
		 */
		u16 get_flip_v() const;
		
		/*! \brief flip sprite horizontally
		 */
		void flip_h();
		
		/*!	\brief set sprite horizontal flip
			\param flip 0 if no flip, non-0 value if flipped
		 */
		void set_flip_h(u16 flip);
		
		/*! \brief get sprite horizontal flip
			\return 0 if no flip, 1 if flipped
		 */
		u16 get_flip_h() const;
		
		/*! \brief set sprite X coordinate
			\param x on-screen sprite X-coordinate (9 bits)
		 */
		void set_x(s16 x);
		
		/*! \brief set sprite Y coordinate
			\param y on-screen sprite Y-coordinate (8 bits)
		 */
		void set_y(s16 y);
		
		/*! \brief get sprite X coordinate
			\return on-screen sprite X coordinate
		 */
		s16 get_x() const;
		
		/*! \brief get sprite Y coordinate
			\return on-screen sprite Y coordinate
		 */
		s16 get_y() const;
		
		/*! \brief set sprite priority
			\param priority sprite priority (0..3)
		 */
		void set_priority(u16 priority);
		
		/*! \brief set index of the first sprite tile 
			\param tile_index sprite tile index
		 */
		void set_tile_index(u16 tile_index);	
		
		/*! \brief gets tile index of the first sprite tile			
			\returns object tile index
		 */
		u16 get_tile_index() const;
		
		/*! \brief makes object visible on the screen
		 */
		void show();
		
		/*! \brief makes object invisible on the screen
		 */
		void hide();

		/*! \brief checks whether object is visible or not
			\returns true if object is hidden, false otherwise
		 */
		bool is_hidden() const;
		
		/*! \brief checks if object is 8-bit
			\returns true if object is 8-bit, false otherwise
		 */
		bool is_8bit() const;
		
		/*! \brief checks if object is 4-bit
			\returns true if object is 4-bit, false otherwise
		 */
		bool is_4bit() const;
		
		int get_palette_number() const;
		
		void set_palette_number(int slot);
		
		bool is_rotation_scaling() const;
		void set_rotation_scaling(bool value);
		
		bool is_double_size() const;
		void set_double_size(bool value);		
		
		
		void set_affine_matrix(int matrix_id);
		int get_affine_matrix() const;
	};
	
	/*! \brief Low level OAM handler
	* 
	 */
	namespace OamPool
	{			
		/*!
		* \brief sets OAM to its original state
		 */
		void reset();
		
		/*!
		* \brief Adds new object to OAM
		* 
		* \param [in] attr new object attributes
		* \return An identificator pointer of the object
		* 
		* \details **Caution!** Use the returned pointer attribute to work with the object. Do not attempt to free the pointer.
		*	 	Use remove_obj() to safely dispose of the object.
		* \details This function operates on a shadow memory zone. Use OamPool::deploy() to make the changes visible on the screen.
		 */
		ObjAttribute* add_obj(ObjAttribute attr);	
		
		/*!
		* \brief Removes object from OAM by its identificator pointer
		* 
		* \param [in] attr pointer to object's attributes
		* 
		* \details **Caution!** Use only the pointers obtained from add_obj() as parameters in this function.
		* \details This function operates on a shadow memory zone. Use OamPool::deploy() to make the changes visible on the screen.
		 */
		void remove_obj(ObjAttribute* attr);
		
		/*!
		* \brief Updates all objects on the screen according to their OAM state.
		 */
		void deploy();
		
		void set_rotation_matrix(int id, int pa, int pb, int pc, int pd);
	}
}
