/*!
* \file fixed_vector.hpp
* \brief Fixed size generic vector definition
 */
 
#pragma once

#include "AstralbrewEngine/types/templates.hpp"
#include "AstralbrewEngine/debug/error.hpp"
#include "AstralbrewEngine/debug/callstack.hpp"

namespace Astralbrew
{
	/*!
	* \brief Fixed size generic vector 
	* \tparam T type of contained data
	* \tparam C vector capacity
	 */
	template<typename T, unsigned int C> class FixedVector
	{
	private:		
		T* container;
		static const uint_best_fit<C> capacity = C;
	public:				
		FixedVector();			
		FixedVector(const FixedVector<T,C>& vector);
		FixedVector(FixedVector<T,C>&& vector);
				
		FixedVector<T,C>& operator = (FixedVector<T,C>&& vector);
		FixedVector<T,C>& operator = (const FixedVector<T,C>& vector);
		
		/*!
		* \brief Finds an item in the vector
		* 
		* \param [in] item Item to be found
		* \return true if the item was found, false otherwise		
		 */
		bool find(const T& item) const;				
		
		/*!
		* \brief Random access index operator
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position	
		* 		
		* \details Range checks are performed only in debug mode.
		 */
		T& operator[] (unsigned int index);		
		
		/*!
		* \brief Random access index operator
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position		
		* 
		* \details Can be used as an lvalue e.g. `v[i] = my_T_Element;`
		* \details Range checks are performed only in debug mode.
		 */
		const T& operator[] (unsigned int index) const;		
		
		/*!
		* \brief Gets element at a certain index
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position		
		* 
		* \details Range checks are performed only in debug mode.
		 */
		const T& get_at(unsigned int index) const;
		
		/*!
		* \brief Gets the size of the vector
		* 
		* \return Vector capacity
		* 
		* \details Being a fixed vector, its size remains constant during its lifespan
		 */
		static uint_best_fit<C> size() { return C; }
		
		~FixedVector();
	};
	
	template<typename T, unsigned int C>
	FixedVector<T,C>::FixedVector()
	{
		container = new T[C];
	}		
	
	template<typename T, unsigned int C>
	FixedVector<T,C>::FixedVector(const FixedVector<T,C>& vector)
	{
		dmaCopy(vector.container, container, C*sizeof(T));
	}
	
	template<typename T, unsigned int C>
	FixedVector<T,C>::FixedVector(FixedVector<T,C>&& vector)
	{
		container = vector.container;
		vector.container = NULL;
	}
	
	template<typename T, unsigned int C>
	FixedVector<T,C>& FixedVector<T,C>::operator = (FixedVector<T,C>&& vector)
	{
		container = vector.container;
		vector.container = NULL;
		return *this;
	}
	
	template<typename T, unsigned int C>
	FixedVector<T,C>& FixedVector<T,C>::operator = (const FixedVector<T,C>& vector)
	{
		dmaCopy(vector.container, container, C*sizeof(T));
		return *this;
	}
	
	template<typename T, unsigned int C>
	bool FixedVector<T,C>::find(const T& item) const
	{
		for(u32 i=0;i<C;i++)
		{
			if(container[i]==item)
				return true;
		}
		return false;
	}
	
	template<typename T, unsigned int C>	
	T& FixedVector<T,C>::operator[] (unsigned int index)
	{
		callstack_call
		if(index>C)
			FATAL_ERROR("Index out of range");
		callstack_ret container[index];
	}
	
	template<typename T, unsigned int C>	
	const T& FixedVector<T,C>::operator[] (unsigned int index) const
	{
		callstack_call
		if(index>C)
			FATAL_ERROR("Index out of range");
		callstack_ret container[index];
	}
	
	template<typename T, unsigned int C>
	const T& FixedVector<T,C>::get_at(unsigned int index) const
	{
		callstack_call
		if(index>C)
			FATAL_ERROR("Index out of range");
		callstack_ret container[index];
	}
	
	template<typename T, unsigned int C>
	FixedVector<T,C>::~FixedVector()
	{
		delete[] container;
	}
	
}



