/*!
* \file vector.hpp
* \brief Generic dynamic vector definition
 */
 
#pragma once

#include <gba.h>
#include "AstralbrewEngine/debug/callstack.hpp"
#include "AstralbrewEngine/debug/error.hpp"

namespace Astralbrew
{
	/*!
	* \brief Generic dynamic vector
	* \tparam T type of contained items
	 */
	template<typename T> class Vector
	{
	private:
		T *container;
		int _size;
		int capacity;
		void container_resize(int new_cap);
	public:				
		Vector(int cap=1);
		Vector(const Vector<T>& vector);
		Vector(Vector<T>&& vector);
		
		Vector<T>& operator = (Vector<T>&& vector);
		Vector<T>& operator = (const Vector<T>& vector);
		
		/*!
		* \brief Sets all the elements in the vector to their default value determined by their type
		* \details Not to be confused with reset(), which also sets the vector to its initials size
		 */
		void clear();
		
		/*!
		* \brief Sets the vector to its initial state		
		 */
		void reset();
		
		
		/*!
		* \brief Changes vector's number of elements
		* 
		* \param [in] new_cap New vector size		
		* 
		* \details If the vector becomes smaller after this operation, the left-out elements are discarded
		 */
		void resize(int new_size);
		
		/*!
		* \brief Adds new element to the end of the vector
		* 
		* \param [in] item Element to be added				
		 */
		void push_back(const T& item);
		
		/*!
		* \brief Gets the last element in vector
		* 
		* \param [output] The element on the last position in vector 
		* \details Error is raised is the vector is empty
		 */
		T& back();
		
		/*!
		* \brief Checks if an item exists in the vector
		* 
		* \param [in] item Item to be found
		* \return true if the item was found, false otherwise		
		 */
		bool exists(const T& item) const;				
		
		
		/*!
		* \brief Finds an item in the vector
		* 
		* \param [in] item Item to be found
		* \return position of the first occurent if element appears in the vector, -1 otherwise
		 */
		int find(const T& item) const;
		
		/*!
		* \brief Removes the first occurence of an element from the vector
		* 
		* \param [in] item Element to be removed
		* \return true if removal was successful, false otherwise (item not found)		
		 */
		bool remove(const T& item);		
		
		/*!
		* \brief Removes the element at a given position
		* 
		* \param [in] index Position of the element to remove		
		* 
		* \details Range checks are performed only in debug mode.
		 */
		void remove_at(int index);
				
		/*!
		* \brief Gets the vector length
		* 
		* \return Number of elements in vector		
		 */
		int size() const;
		
		/*!
		* \brief Random access index operator
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position	
		* 		
		* \details Range checks are performed only in debug mode.
		 */
		T& operator[] (int index);		
		
		/*!
		* \brief Random access index operator
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position		
		* 
		* \details Can be used as an lvalue e.g. `v[i] = my_T_Element;`
		* \details Range checks are performed only in debug mode.
		 */
		const T& operator[] (int index) const;		
		
		/*!
		* \brief Gets element at a certain index
		* 
		* \param [in] index Position in vector
		* \return Element at the specified position		
		* 
		* \details Range checks are performed only in debug mode.
		 */
		const T& get_at(int index) const;
		
		~Vector();
	};
}

template<typename T>
Astralbrew::Vector<T>::Vector(int cap)
{	
	_size = cap;
	if(sizeof(T)&1)
	{
		cap>>=1; cap++; cap<<=1; // assure capacity is 16-bit aligned
	}
	capacity = cap;	
	container = new T[cap];		
	clear();
}

template<typename T>
Astralbrew::Vector<T>::Vector(Vector<T>&& vector)
{
	_size = vector._size; 	
	capacity = vector.capacity;
	container = vector.container;
	vector.container = NULL;		
	vector._size = 0;
	vector.capacity = 0;
}

template<typename T>
Astralbrew::Vector<T>& Astralbrew::Vector<T>::operator = (Astralbrew::Vector<T>&& vector)
{
	_size = vector.size();
	capacity = vector.capacity;
	container = vector.container;
	vector.container = NULL;	
	vector._size = 0;
	vector.capacity = 0;
	return *this;
}

template<typename T>
Astralbrew::Vector<T>::Vector(const Vector<T>& vector)
{
	_size = vector._size;
	capacity = vector.capacity;
	delete[] container;
	container = new T[capacity];	
	dmaCopy(vector.container, container, _size * sizeof(T));	
}

template<typename T>
Astralbrew::Vector<T>& Astralbrew::Vector<T>::operator = (const Astralbrew::Vector<T>& vector)
{
	_size = vector._size;
	capacity = vector.capacity;
	delete[] container;
	container = new T[capacity];	
	dmaCopy(vector.container, container, _size * sizeof(T));	
	return *this;
}

template<typename T>
void Astralbrew::Vector<T>::clear()
{
	_size=0;
}

template<typename T>
void Astralbrew::Vector<T>::reset()
{
	container_resize(1);
	_size = 0;
}

template<typename T>
void Astralbrew::Vector<T>:: container_resize(int new_cap)
{
	if(sizeof(T)&1)
	{
		new_cap>>=1; new_cap++; new_cap<<=1;
	}
	T* new_container = new T[new_cap];
	int min_cap = new_cap < _size ? new_cap : _size;
	if(min_cap>0)
	{
		//dmaCopy(container, new_container, min_cap * sizeof(T));
		for(int i=0;i<min_cap;i++)
		{
			new_container[i] = (T&&)container[i];
		}
	}
	delete[] container;
	container = new_container;
	capacity = new_cap;
}

template<typename T>
void Astralbrew::Vector<T>::resize(int new_size)
{		
	container_resize(new_size);
	_size = new_size;
}

template<typename T>
void Astralbrew::Vector<T>::push_back(const T& item)
{
	if(_size==capacity)
	{
		container_resize(capacity<<1);
	}
	container[_size++]=item;
}

template<typename T>
T& Astralbrew::Vector<T>::operator[] (int index)
{
	callstack_call
	#ifdef VERSION_DEBUG
	if(index>=_size)
		FATAL_ERROR("Index out of range");
	#endif
	
	callstack_ret container[index];
}

template<typename T>
const T& Astralbrew::Vector<T>::operator[] (int index) const
{
	callstack_call
	#ifdef VERSION_DEBUG
	if(index>=_size)
		FATAL_ERROR("Index out of range");
	#endif
	
	callstack_ret container[index];
}

template<typename T>
const T& Astralbrew::Vector<T>::get_at(int index) const
{
	callstack_call
	
	#ifdef VERSION_DEBUG
	if(index>=_size)
		FATAL_ERROR("Index out of range");
	#endif
	
	callstack_ret container[index];
}

template<typename T>
bool Astralbrew::Vector<T>::exists(const T& item) const
{
	for(int i=0;i<_size;i++)
	{
		if(container[i]==item)
			return true;
	}
	return false;
}

template<typename T>
int Astralbrew::Vector<T>::find(const T& item) const
{
	for(int i=0;i<_size;i++)
	{
		if(container[i]==item)
			return i;
	}
	return -1;
}

template<typename T>
bool Astralbrew::Vector<T>::remove(const T& item)
{
	for(int i=0;i<_size;i++)
	{
		if(container[i]==item)
		{
			_size--;
			for(int j=i;j<_size;j++)
			{
				container[j] = container[j+1];
			}
			return true;
		}
	}
	return false;
}

template<typename T>
void Astralbrew::Vector<T>::remove_at(int index)
{
	callstack_call
	#ifdef VERSION_DEBUG
	if(index >= _size)
		FATAL_ERROR("Index out of range");
	#endif
	_size--;
	for(int j=index;j<_size;j++)
	{
		container[j] = container[j+1];
	}	
	callstack_ret;
}

template<typename T>
int Astralbrew::Vector<T>::size() const
{
	return _size;
}

template<typename T>
Astralbrew::Vector<T>::~Vector()
{
	delete[] container;
}

template<typename T>
T& Astralbrew::Vector<T>::back()
{
	callstack_call
	assert(size()>0);
	callstack_ret container[size()-1];
}
